__author__ = 'Martin'

class ArrayError(Exception):
    pass

class array_(list):
    def __init__(self, *dims, numbered = False, fill = None):
        super(array_, self).__init__()
        self.dims = dims
        self.ndims = len(dims)
        self.mults = [1]*self.ndims
        self.size = 1
        ix = self.ndims - 1
        while ix > 0:
            self.size *= dims[ix]
            self.mults[ix-1] = self.mults[ix] * dims[ix]
            ix -= 1
        self.size *= dims[0]

        if numbered:
            ext = [n for n in range(self.size)]
            self.extend(ext)
        else:
            self.extend([fill]*self.size)
        pass

    def _compute_index(self, indices):
        if type(indices) is not tuple:
            return indices
        if len(indices) != self.ndims:
            raise IndexError()
        ix = 0
        for i in range(self.ndims):
            ix += indices[i] * self.mults[i]
        return ix

    def __getitem__(self, indices):
        ix = self._compute_index(indices)
        return super().__getitem__(ix)

    def __setitem__(self, indices, value):
        ix = self._compute_index(indices)
        super().__setitem__(ix, value)

    # exclude list methods not provided by basic arrays
    def append(self, item):
        raise ArrayError('Arrays are not extendable')
    def pop(index=None):
        raise ArrayError('pop() not implemented for array')
    def remove(self, value):
        raise ArrayError('remove() not implemented for array')

    def assign(self, sequence):
        if len(sequence) != self.size:
            raise IndexError()
        self.clear()
        self.extend(sequence)

    def _getstr(self, level, lastdim, indices, result):
        """
        modifies successive copies of a list of indices starting with [0, 0, ...]  to index elements
        __getitem__ is called with this list to get the corresponding element
        :param level:   loop nesting level 0..lastdim
        :param lastdim: index to the last dimension in self.dims
        :param indices: current list of indices
        :param result:  partial string representation of the array on entry
        :return:        partial or complete string representation of the array on exit
        """
        local_inds = list(indices)     # create a copy
        if level == lastdim:
            for i in range(self.dims[level]):       # indices to elements in current vector
                local_inds[level] = i               # modify local dims at innermost level
                result += str(self.__getitem__(tuple(local_inds))) + ', '
        else:
            for i in range(self.dims[level]):
                result = self._getstr(level+1, lastdim, local_inds, result)
                local_inds[level] += 1              # modify local dims at outer level
                if result[-4:] != '\n \n ':
                    result += '\n '
        return result

    def as_str(self):
        if self.ndims == 1:
            return super().__str__()
        lastdim = self.ndims - 1
        level = 0
        result = '['
        indices = [0]*self.ndims    # one index per dimension
        result =  self._getstr(level, lastdim, indices, result)
        return result[:result.rindex(',')] +']'

    def __str__(self):
        return self.as_str()

def copy_right(source, s_index, dest, d_index, length):
    sx = s_index + length - 1
    dx = d_index + length - 1
    for sx in range(sx, s_index - 1, -1):
        dest[dx] = source[sx]
        dx -= 1

def copy_left(source, s_index, dest, d_index, length):
    for sx in range(s_index, s_index + length):
        dest[d_index] = source[sx]
        d_index += 1

def array_cpy(source, s_index, dest, d_index, length):
    if source is dest and s_index < d_index:
        sx = s_index + length - 1
        dx = d_index + length - 1
        for sx in range(sx, s_index - 1, -1):
            dest[dx] = source[sx]
            dx -= 1
    else:
        for sx in range(s_index, s_index + length):
            dest[d_index] = source[sx]
            d_index += 1



if __name__ == '__main__':

    ary = array_(10, numbered=True)
    ary[9] = '#'
    print(ary)
    copy_right(ary, 1, ary, 2, 8)
    print(ary)
    ary[1] = '#'
    print(ary)
    copy_left(ary, 2, ary, 1, 8)
    print(ary)
    print()

    ary = array_(10, numbered=True)
    ary[9] = '#'
    print(ary)
    array_cpy(ary, 1, ary, 2, 8)
    print(ary)
    ary[1] = '#'
    print(ary)
    array_cpy(ary, 2, ary, 1, 8)
    print(ary)
    print()

    ary = array_(3, 3, 3, numbered=True)
    print(ary)